from django import forms
from django.core import exceptions
from django.db import models
from django.test import SimpleTestCase
from django.test.utils import modify_settings
from forms_tests.widget_tests.base import WidgetTest

from django_mongodb_backend.fields import ArrayField
from django_mongodb_backend.forms import SimpleArrayField, SplitArrayField, SplitArrayWidget


class IntegerArrayModel(models.Model):
    field = ArrayField(models.IntegerField(), default=list, blank=True)


class SimpleArrayFieldTests(SimpleTestCase):
    def test_valid(self):
        field = SimpleArrayField(forms.CharField())
        value = field.clean("a,b,c")
        self.assertEqual(value, ["a", "b", "c"])

    def test_to_python_fail(self):
        field = SimpleArrayField(forms.IntegerField())
        msg = "Item 1 in the array did not validate: Enter a whole number."
        with self.assertRaisesMessage(exceptions.ValidationError, msg):
            field.clean("a,b,9")

    def test_validate_fail(self):
        field = SimpleArrayField(forms.CharField(required=True))
        msg = "Item 3 in the array did not validate: This field is required."
        with self.assertRaisesMessage(exceptions.ValidationError, msg):
            field.clean("a,b,")

    def test_validate_fail_base_field_error_params(self):
        field = SimpleArrayField(forms.CharField(max_length=2))
        with self.assertRaises(exceptions.ValidationError) as cm:
            field.clean("abc,c,defg")
        errors = cm.exception.error_list
        self.assertEqual(len(errors), 2)
        first_error = errors[0]
        self.assertEqual(
            first_error.message,
            "Item 1 in the array did not validate: Ensure this value has at most 2 "
            "characters (it has 3).",
        )
        self.assertEqual(first_error.code, "item_invalid")
        self.assertEqual(
            first_error.params,
            {"nth": 1, "value": "abc", "limit_value": 2, "show_value": 3},
        )
        second_error = errors[1]
        self.assertEqual(
            second_error.message,
            "Item 3 in the array did not validate: Ensure this value has at most 2 "
            "characters (it has 4).",
        )
        self.assertEqual(second_error.code, "item_invalid")
        self.assertEqual(
            second_error.params,
            {"nth": 3, "value": "defg", "limit_value": 2, "show_value": 4},
        )

    def test_validators_fail(self):
        field = SimpleArrayField(forms.RegexField("[a-e]{2}"))
        msg = "Item 1 in the array did not validate: Enter a valid value."
        with self.assertRaisesMessage(exceptions.ValidationError, msg):
            field.clean("a,bc,de")

    def test_delimiter(self):
        field = SimpleArrayField(forms.CharField(), delimiter="|")
        value = field.clean("a|b|c")
        self.assertEqual(value, ["a", "b", "c"])

    def test_delimiter_with_nesting(self):
        field = SimpleArrayField(SimpleArrayField(forms.CharField()), delimiter="|")
        value = field.clean("a,b|c,d")
        self.assertEqual(value, [["a", "b"], ["c", "d"]])

    def test_prepare_value(self):
        field = SimpleArrayField(forms.CharField())
        value = field.prepare_value(["a", "b", "c"])
        self.assertEqual(value, "a,b,c")

    def test_max_length(self):
        field = SimpleArrayField(forms.CharField(), max_length=2)
        msg = "List contains 3 items, it should contain no more than 2."
        with self.assertRaisesMessage(exceptions.ValidationError, msg):
            field.clean("a,b,c")

    def test_min_length(self):
        field = SimpleArrayField(forms.CharField(), min_length=4)
        msg = "List contains 3 items, it should contain no fewer than 4."
        with self.assertRaisesMessage(exceptions.ValidationError, msg):
            field.clean("a,b,c")

    def test_min_length_singular(self):
        field = SimpleArrayField(forms.IntegerField(), min_length=2)
        field.clean([1, 2])
        msg = "List contains 1 item, it should contain no fewer than 2."
        with self.assertRaisesMessage(exceptions.ValidationError, msg):
            field.clean([1])

    def test_size_length(self):
        field = SimpleArrayField(forms.CharField(max_length=27), length=4)
        msg = "List contains 3 items, it should contain 4."
        with self.assertRaisesMessage(exceptions.ValidationError, msg):
            field.clean(["a", "b", "c"])
        msg = "List contains 5 items, it should contain 4."
        with self.assertRaisesMessage(exceptions.ValidationError, msg):
            field.clean(["a", "b", "c", "d", "e"])

    def test_size_length_singular(self):
        field = SimpleArrayField(forms.CharField(max_length=27), length=4)
        msg = "List contains 1 item, it should contain 4."
        with self.assertRaisesMessage(exceptions.ValidationError, msg):
            field.clean(["a"])

    def test_required(self):
        field = SimpleArrayField(forms.CharField(), required=True)
        msg = "This field is required."
        with self.assertRaisesMessage(exceptions.ValidationError, msg):
            field.clean("")

    def test_length_and_max_min_length(self):
        msg = "The length and max_length parameters are mutually exclusive."
        with self.assertRaisesMessage(exceptions.ImproperlyConfigured, msg):
            SimpleArrayField(forms.CharField(), max_length=3, length=2)
        msg = "The length and min_length parameters are mutually exclusive."
        with self.assertRaisesMessage(exceptions.ImproperlyConfigured, msg):
            SimpleArrayField(forms.CharField(), min_length=3, length=2)

    def test_model_field_formfield(self):
        model_field = ArrayField(models.CharField(max_length=27))
        form_field = model_field.formfield()
        self.assertIsInstance(form_field, SimpleArrayField)
        self.assertIsInstance(form_field.base_field, forms.CharField)
        self.assertEqual(form_field.base_field.max_length, 27)

    def test_model_field_formfield_max_size(self):
        model_field = ArrayField(models.CharField(max_length=27), max_size=4)
        form_field = model_field.formfield()
        self.assertIsInstance(form_field, SimpleArrayField)
        self.assertEqual(form_field.max_length, 4)

    def test_model_field_formfield_size(self):
        model_field = ArrayField(models.CharField(max_length=27), size=4)
        form_field = model_field.formfield()
        self.assertIsInstance(form_field, SimpleArrayField)
        self.assertEqual(form_field.length, 4)

    def test_model_field_choices(self):
        model_field = ArrayField(models.IntegerField(choices=((1, "A"), (2, "B"))))
        form_field = model_field.formfield()
        self.assertEqual(form_field.clean("1,2"), [1, 2])

    def test_already_converted_value(self):
        field = SimpleArrayField(forms.CharField())
        vals = ["a", "b", "c"]
        self.assertEqual(field.clean(vals), vals)

    def test_has_changed(self):
        field = SimpleArrayField(forms.IntegerField())
        self.assertIs(field.has_changed([1, 2], [1, 2]), False)
        self.assertIs(field.has_changed([1, 2], "1,2"), False)
        self.assertIs(field.has_changed([1, 2], "1,2,3"), True)
        self.assertIs(field.has_changed([1, 2], "a,b"), True)

    def test_has_changed_empty(self):
        field = SimpleArrayField(forms.CharField())
        self.assertIs(field.has_changed(None, None), False)
        self.assertIs(field.has_changed(None, ""), False)
        self.assertIs(field.has_changed(None, []), False)
        self.assertIs(field.has_changed([], None), False)
        self.assertIs(field.has_changed([], ""), False)


# To locate the widget's template.
@modify_settings(INSTALLED_APPS={"append": "django_mongodb_backend"})
class SplitFormFieldTests(SimpleTestCase):
    def test_valid(self):
        class SplitForm(forms.Form):
            array = SplitArrayField(forms.CharField(), size=3)

        data = {"array_0": "a", "array_1": "b", "array_2": "c"}
        form = SplitForm(data)
        self.assertTrue(form.is_valid())
        self.assertEqual(form.cleaned_data, {"array": ["a", "b", "c"]})

    def test_required(self):
        class SplitForm(forms.Form):
            array = SplitArrayField(forms.CharField(), required=True, size=3)

        data = {"array_0": "", "array_1": "", "array_2": ""}
        form = SplitForm(data)
        self.assertFalse(form.is_valid())
        self.assertEqual(form.errors, {"array": ["This field is required."]})

    def test_remove_trailing_nulls(self):
        class SplitForm(forms.Form):
            array = SplitArrayField(
                forms.CharField(required=False), size=5, remove_trailing_nulls=True
            )

        data = {
            "array_0": "a",
            "array_1": "",
            "array_2": "b",
            "array_3": "",
            "array_4": "",
        }
        form = SplitForm(data)
        self.assertTrue(form.is_valid(), form.errors)
        self.assertEqual(form.cleaned_data, {"array": ["a", "", "b"]})

    def test_remove_trailing_nulls_not_required(self):
        class SplitForm(forms.Form):
            array = SplitArrayField(
                forms.CharField(required=False),
                size=2,
                remove_trailing_nulls=True,
                required=False,
            )

        data = {"array_0": "", "array_1": ""}
        form = SplitForm(data)
        self.assertTrue(form.is_valid())
        self.assertEqual(form.cleaned_data, {"array": []})

    def test_required_field(self):
        class SplitForm(forms.Form):
            array = SplitArrayField(forms.CharField(), size=3)

        data = {"array_0": "a", "array_1": "b", "array_2": ""}
        form = SplitForm(data)
        self.assertFalse(form.is_valid())
        self.assertEqual(
            form.errors,
            {"array": ["Item 3 in the array did not validate: This field is required."]},
        )

    def test_invalid_integer(self):
        msg = (
            "Item 2 in the array did not validate: Ensure this value is less than or "
            "equal to 100."
        )
        with self.assertRaisesMessage(exceptions.ValidationError, msg):
            SplitArrayField(forms.IntegerField(max_value=100), size=2).clean([0, 101])

    def test_rendering(self):
        class SplitForm(forms.Form):
            array = SplitArrayField(forms.CharField(), size=3)

        self.assertHTMLEqual(
            str(SplitForm()),
            """
            <div>
                <label for="id_array_0">Array:</label>
                <input id="id_array_0" name="array_0" type="text" required>
                <input id="id_array_1" name="array_1" type="text" required>
                <input id="id_array_2" name="array_2" type="text" required>
            </div>
        """,
        )

    def test_invalid_char_length(self):
        field = SplitArrayField(forms.CharField(max_length=2), size=3)
        with self.assertRaises(exceptions.ValidationError) as cm:
            field.clean(["abc", "c", "defg"])
        self.assertEqual(
            cm.exception.messages,
            [
                "Item 1 in the array did not validate: Ensure this value has at most 2 "
                "characters (it has 3).",
                "Item 3 in the array did not validate: Ensure this value has at most 2 "
                "characters (it has 4).",
            ],
        )

    def test_splitarraywidget_value_omitted_from_data(self):
        class Form(forms.ModelForm):
            field = SplitArrayField(forms.IntegerField(), required=False, size=2)

            class Meta:
                model = IntegerArrayModel
                fields = ("field",)

        form = Form({"field_0": "1", "field_1": "2"})
        self.assertEqual(form.errors, {})
        obj = form.save(commit=False)
        self.assertEqual(obj.field, [1, 2])

    def test_splitarrayfield_has_changed(self):
        class Form(forms.ModelForm):
            field = SplitArrayField(forms.IntegerField(), required=False, size=2)

            class Meta:
                model = IntegerArrayModel
                fields = ("field",)

        tests = [
            ({}, {"field_0": "", "field_1": ""}, True),
            ({"field": None}, {"field_0": "", "field_1": ""}, True),
            ({"field": [1]}, {"field_0": "", "field_1": ""}, True),
            ({"field": [1]}, {"field_0": "1", "field_1": "0"}, True),
            ({"field": [1, 2]}, {"field_0": "1", "field_1": "2"}, False),
            ({"field": [1, 2]}, {"field_0": "a", "field_1": "b"}, True),
        ]
        for initial, data, expected_result in tests:
            with self.subTest(initial=initial, data=data):
                obj = IntegerArrayModel(**initial)
                form = Form(data, instance=obj)
                self.assertIs(form.has_changed(), expected_result)

    def test_splitarrayfield_remove_trailing_nulls_has_changed(self):
        class Form(forms.ModelForm):
            field = SplitArrayField(
                forms.IntegerField(), required=False, size=2, remove_trailing_nulls=True
            )

            class Meta:
                model = IntegerArrayModel
                fields = ("field",)

        tests = [
            ({}, {"field_0": "", "field_1": ""}, False),
            ({"field": None}, {"field_0": "", "field_1": ""}, False),
            ({"field": []}, {"field_0": "", "field_1": ""}, False),
            ({"field": [1]}, {"field_0": "1", "field_1": ""}, False),
        ]
        for initial, data, expected_result in tests:
            with self.subTest(initial=initial, data=data):
                obj = IntegerArrayModel(**initial)
                form = Form(data, instance=obj)
                self.assertIs(form.has_changed(), expected_result)


# To locate the widget's template.
@modify_settings(INSTALLED_APPS={"append": "django_mongodb_backend"})
class SplitArrayWidgetTests(WidgetTest, SimpleTestCase):
    def test_get_context(self):
        self.assertEqual(
            SplitArrayWidget(forms.TextInput(), size=2).get_context("name", ["val1", "val2"]),
            {
                "widget": {
                    "name": "name",
                    "is_hidden": False,
                    "required": False,
                    "value": "['val1', 'val2']",
                    "attrs": {},
                    "template_name": "mongodb/widgets/split_array.html",
                    "subwidgets": [
                        {
                            "name": "name_0",
                            "is_hidden": False,
                            "required": False,
                            "value": "val1",
                            "attrs": {},
                            "template_name": "django/forms/widgets/text.html",
                            "type": "text",
                        },
                        {
                            "name": "name_1",
                            "is_hidden": False,
                            "required": False,
                            "value": "val2",
                            "attrs": {},
                            "template_name": "django/forms/widgets/text.html",
                            "type": "text",
                        },
                    ],
                }
            },
        )

    def test_checkbox_get_context_attrs(self):
        context = SplitArrayWidget(
            forms.CheckboxInput(),
            size=2,
        ).get_context("name", [True, False])
        self.assertEqual(context["widget"]["value"], "[True, False]")
        self.assertEqual(
            [subwidget["attrs"] for subwidget in context["widget"]["subwidgets"]],
            [{"checked": True}, {}],
        )

    def test_render(self):
        self.check_html(
            SplitArrayWidget(forms.TextInput(), size=2),
            "array",
            None,
            """
            <input name="array_0" type="text">
            <input name="array_1" type="text">
            """,
        )

    def test_render_attrs(self):
        self.check_html(
            SplitArrayWidget(forms.TextInput(), size=2),
            "array",
            ["val1", "val2"],
            attrs={"id": "foo"},
            html=(
                """
                <input id="foo_0" name="array_0" type="text" value="val1">
                <input id="foo_1" name="array_1" type="text" value="val2">
                """
            ),
        )

    def test_value_omitted_from_data(self):
        widget = SplitArrayWidget(forms.TextInput(), size=2)
        self.assertIs(widget.value_omitted_from_data({}, {}, "field"), True)
        self.assertIs(widget.value_omitted_from_data({"field_0": "value"}, {}, "field"), False)
        self.assertIs(widget.value_omitted_from_data({"field_1": "value"}, {}, "field"), False)
        self.assertIs(
            widget.value_omitted_from_data({"field_0": "value", "field_1": "value"}, {}, "field"),
            False,
        )



from django.conf import settings
from django.db.backends.base.creation import BaseDatabaseCreation

from django_mongodb_backend.management.commands.createcachecollection import (
    Command as CreateCacheCollection,
)


class DatabaseCreation(BaseDatabaseCreation):
    def _execute_create_test_db(self, cursor, parameters, keepdb=False):
        if not keepdb:
            self._destroy_test_db(parameters["dbname"], verbosity=0)

    def _destroy_test_db(self, test_database_name, verbosity):
        # At this point, settings still points to the non-test database. For
        # MongoDB, it must use the test database.
        settings.DATABASES[self.connection.alias]["NAME"] = test_database_name
        self.connection.settings_dict["NAME"] = test_database_name

        for collection in self.connection.introspection.table_names():
            if not collection.startswith("system."):
                self.connection.database.drop_collection(collection)

    def create_test_db(self, *args, **kwargs):
        test_database_name = super().create_test_db(*args, **kwargs)
        # Not using call_command() avoids the requirement to put
        # "django_mongodb_backend" in INSTALLED_APPS.
        CreateCacheCollection().handle(
            database=self.connection.alias, verbosity=kwargs["verbosity"]
        )
        return test_database_name



import itertools

from django.db import connection, models
from django.test import TransactionTestCase
from django.test.utils import isolate_apps

from django_mongodb_backend.fields import EmbeddedModelField
from django_mongodb_backend.models import EmbeddedModel

from .models import Address, Author, Book, new_apps


class TestMixin:
    available_apps = []
    models = [Address, Author, Book]

    # Utility functions

    def setUp(self):
        # local_models should contain test dependent model classes that will be
        # automatically removed from the app cache on test tear down.
        self.local_models = []
        # isolated_local_models contains models that are in test methods
        # decorated with @isolate_apps.
        self.isolated_local_models = []

    def tearDown(self):
        # Delete any tables made for our models
        self.delete_tables()
        new_apps.clear_cache()
        for model in new_apps.get_models():
            model._meta._expire_cache()
        if "schema" in new_apps.all_models:
            for model in self.local_models:
                for many_to_many in model._meta.many_to_many:
                    through = many_to_many.remote_field.through
                    if through and through._meta.auto_created:
                        del new_apps.all_models["schema"][through._meta.model_name]
                del new_apps.all_models["schema"][model._meta.model_name]
        if self.isolated_local_models:
            with connection.schema_editor() as editor:
                for model in self.isolated_local_models:
                    editor.delete_model(model)

    def delete_tables(self):
        "Deletes all model tables for our models for a clean test environment"
        converter = connection.introspection.identifier_converter
        with connection.schema_editor() as editor:
            connection.disable_constraint_checking()
            table_names = connection.introspection.table_names()
            if connection.features.ignores_table_name_case:
                table_names = [table_name.lower() for table_name in table_names]
            for model in itertools.chain(SchemaTests.models, self.local_models):
                tbl = converter(model._meta.db_table)
                if connection.features.ignores_table_name_case:
                    tbl = tbl.lower()
                if tbl in table_names:
                    editor.delete_model(model)
                    table_names.remove(tbl)
            connection.enable_constraint_checking()

    def get_constraints(self, table):
        """
        Get the constraints on a table using a new cursor.
        """
        with connection.cursor() as cursor:
            return connection.introspection.get_constraints(cursor, table)

    def get_constraints_for_columns(self, model, columns):
        constraints = self.get_constraints(model._meta.db_table)
        constraints_for_column = []
        for name, details in constraints.items():
            if details["columns"] == columns:
                constraints_for_column.append(name)
        return sorted(constraints_for_column)

    def assertIndexOrder(self, table, index, order):
        constraints = self.get_constraints(table)
        self.assertIn(index, constraints)
        index_orders = constraints[index]["orders"]
        self.assertTrue(
            all(val == expected for val, expected in zip(index_orders, order, strict=True))
        )

    def assertTableExists(self, model):
        self.assertIn(model._meta.db_table, connection.introspection.table_names())

    def assertTableNotExists(self, model):
        self.assertNotIn(model._meta.db_table, connection.introspection.table_names())


class SchemaTests(TestMixin, TransactionTestCase):
    # SchemaEditor.create_model() tests
    def test_db_index(self):
        """Field(db_index=True) on an embedded model."""
        with connection.schema_editor() as editor:
            # Create the table
            editor.create_model(Book)
            # The table is there
            self.assertTableExists(Book)
            # Embedded indexes are created.
            self.assertEqual(
                self.get_constraints_for_columns(Book, ["author.age"]),
                ["schema__book_author.age_dc08100b"],
            )
            self.assertEqual(
                self.get_constraints_for_columns(Book, ["author.address.zip_code"]),
                ["schema__book_author.address.zip_code_7b9a9307"],
            )
            # Clean up that table
            editor.delete_model(Book)
        # The table is gone
        self.assertTableNotExists(Author)

    def test_unique(self):
        """Field(unique=True) on an embedded model."""
        with connection.schema_editor() as editor:
            editor.create_model(Book)
            self.assertTableExists(Book)
            # Embedded uniques are created.
            self.assertEqual(
                self.get_constraints_for_columns(Book, ["author.employee_id"]),
                ["schema__book_author.employee_id_7d4d3eff_uniq"],
            )
            self.assertEqual(
                self.get_constraints_for_columns(Book, ["author.address.uid"]),
                ["schema__book_author.address.uid_8124a01f_uniq"],
            )
            # Clean up that table
            editor.delete_model(Book)
        self.assertTableNotExists(Author)

    @isolate_apps("schema_")
    def test_unique_together(self):
        """Meta.unique_together on an embedded model."""

        class Address(EmbeddedModel):
            unique_together_one = models.CharField(max_length=10)
            unique_together_two = models.CharField(max_length=10)

            class Meta:
                app_label = "schema_"
                unique_together = [("unique_together_one", "unique_together_two")]

        class Author(EmbeddedModel):
            address = EmbeddedModelField(Address)
            unique_together_three = models.CharField(max_length=10)
            unique_together_four = models.CharField(max_length=10)

            class Meta:
                app_label = "schema_"
                unique_together = [("unique_together_three", "unique_together_four")]

        class Book(models.Model):
            author = EmbeddedModelField(Author)

            class Meta:
                app_label = "schema_"

        with connection.schema_editor() as editor:
            editor.create_model(Book)
            self.assertTableExists(Book)
            # Embedded uniques are created.
            self.assertEqual(
                self.get_constraints_for_columns(
                    Book, ["author.unique_together_three", "author.unique_together_four"]
                ),
                [
                    "schema__author_author.unique_together_three_author.unique_together_four_39e1cb43_uniq"
                ],
            )
            self.assertEqual(
                self.get_constraints_for_columns(
                    Book,
                    ["author.address.unique_together_one", "author.address.unique_together_two"],
                ),
                [
                    "schema__address_author.address.unique_together_one_author.address.unique_together_two_de682e30_uniq"
                ],
            )
            editor.delete_model(Book)
        self.assertTableNotExists(Book)

    @isolate_apps("schema_")
    def test_indexes(self):
        """Meta.indexes on an embedded model."""

        class Address(EmbeddedModel):
            indexed_one = models.CharField(max_length=10)

            class Meta:
                app_label = "schema_"
                indexes = [models.Index(fields=["indexed_one"])]

        class Author(EmbeddedModel):
            address = EmbeddedModelField(Address)
            indexed_two = models.CharField(max_length=10)

            class Meta:
                app_label = "schema_"
                indexes = [models.Index(fields=["indexed_two"])]

        class Book(models.Model):
            author = EmbeddedModelField(Author)

            class Meta:
                app_label = "schema_"

        with connection.schema_editor() as editor:
            editor.create_model(Book)
            self.assertTableExists(Book)
            # Embedded uniques are created.
            self.assertEqual(
                self.get_constraints_for_columns(Book, ["author.indexed_two"]),
                ["schema__aut_indexed_b19137_idx"],
            )
            self.assertEqual(
                self.get_constraints_for_columns(
                    Book,
                    ["author.address.indexed_one"],
                ),
                ["schema__add_indexed_b64972_idx"],
            )
            editor.delete_model(Author)
        self.assertTableNotExists(Author)

    @isolate_apps("schema_")
    def test_constraints(self):
        """Meta.constraints on an embedded model."""

        class Address(models.Model):
            unique_constraint_one = models.CharField(max_length=10)

            class Meta:
                app_label = "schema_"
                constraints = [
                    models.UniqueConstraint(fields=["unique_constraint_one"], name="unique_one")
                ]

        class Author(models.Model):
            address = EmbeddedModelField(Address)
            unique_constraint_two = models.CharField(max_length=10)

            class Meta:
                app_label = "schema_"
                constraints = [
                    models.UniqueConstraint(fields=["unique_constraint_two"], name="unique_two")
                ]

        class Book(models.Model):
            author = EmbeddedModelField(Author)

            class Meta:
                app_label = "schema_"

        with connection.schema_editor() as editor:
            editor.create_model(Book)
            self.assertTableExists(Book)
            # Embedded uniques are created.
            self.assertEqual(
                self.get_constraints_for_columns(Book, ["author.unique_constraint_two"]),
                ["unique_two"],
            )
            self.assertEqual(
                self.get_constraints_for_columns(
                    Book,
                    ["author.address.unique_constraint_one"],
                ),
                ["unique_one"],
            )
            editor.delete_model(Author)
        self.assertTableNotExists(Author)

    # SchemaEditor.add_field() / remove_field() tests
    @isolate_apps("schema_")
    def test_add_remove_field_db_index_and_unique(self):
        """AddField/RemoveField + EmbeddedModelField + Field(db_index=True) & Field(unique=True)."""

        class Book(models.Model):
            name = models.CharField(max_length=100)

            class Meta:
                app_label = "schema_"

        new_field = EmbeddedModelField(Author)
        new_field.set_attributes_from_name("author")
        with connection.schema_editor() as editor:
            # Create the table amd add the field.
            editor.create_model(Book)
            editor.add_field(Book, new_field)
            # Embedded indexes are created.
            self.assertEqual(
                self.get_constraints_for_columns(Book, ["author.age"]),
                ["schema__book_author.age_dc08100b"],
            )
            self.assertEqual(
                self.get_constraints_for_columns(Book, ["author.address.zip_code"]),
                ["schema__book_author.address.zip_code_7b9a9307"],
            )
            # Embedded uniques
            self.assertEqual(
                self.get_constraints_for_columns(Book, ["author.employee_id"]),
                ["schema__book_author.employee_id_7d4d3eff_uniq"],
            )
            self.assertEqual(
                self.get_constraints_for_columns(Book, ["author.address.uid"]),
                ["schema__book_author.address.uid_8124a01f_uniq"],
            )
            editor.remove_field(Book, new_field)
            # Embedded indexes are removed.
            self.assertEqual(
                self.get_constraints_for_columns(Book, ["author.age"]),
                [],
            )
            self.assertEqual(
                self.get_constraints_for_columns(Book, ["author.address.zip_code"]),
                [],
            )
            self.assertEqual(
                self.get_constraints_for_columns(Book, ["author.employee_id"]),
                [],
            )
            self.assertEqual(
                self.get_constraints_for_columns(Book, ["author.address.uid"]),
                [],
            )
            editor.delete_model(Book)
        self.assertTableNotExists(Author)

    @isolate_apps("schema_")
    def test_add_remove_field_unique_together(self):
        """AddField/RemoveField + EmbeddedModelField + Meta.unique_together."""

        class Address(models.Model):
            unique_together_one = models.CharField(max_length=10)
            unique_together_two = models.CharField(max_length=10)

            class Meta:
                app_label = "schema_"
                unique_together = [("unique_together_one", "unique_together_two")]

        class Author(models.Model):
            address = EmbeddedModelField(Address)
            unique_together_three = models.CharField(max_length=10)
            unique_together_four = models.CharField(max_length=10)

            class Meta:
                app_label = "schema_"
                unique_together = [("unique_together_three", "unique_together_four")]

        class Book(models.Model):
            author = EmbeddedModelField(Author)

            class Meta:
                app_label = "schema_"

        new_field = EmbeddedModelField(Author)
        new_field.set_attributes_from_name("author")
        with connection.schema_editor() as editor:
            # Create the table and add the field.
            editor.create_model(Book)
            editor.add_field(Book, new_field)
            # Embedded uniques are created.
            self.assertEqual(
                self.get_constraints_for_columns(
                    Book, ["author.unique_together_three", "author.unique_together_four"]
                ),
                [
                    "schema__author_author.unique_together_three_author.unique_together_four_39e1cb43_uniq"
                ],
            )
            self.assertEqual(
                self.get_constraints_for_columns(
                    Book,
                    ["author.address.unique_together_one", "author.address.unique_together_two"],
                ),
                [
                    "schema__address_author.address.unique_together_one_author.address.unique_together_two_de682e30_uniq"
                ],
            )
            editor.remove_field(Book, new_field)
            # Embedded indexes are removed.
            self.assertEqual(
                self.get_constraints_for_columns(
                    Book, ["author.unique_together_three", "author.unique_together_four"]
                ),
                [],
            )
            self.assertEqual(
                self.get_constraints_for_columns(
                    Book,
                    ["author.address.unique_together_one", "author.address.unique_together_two"],
                ),
                [],
            )
            editor.delete_model(Book)
        self.assertTableNotExists(Book)

    @isolate_apps("schema_")
    def test_add_remove_field_indexes(self):
        """AddField/RemoveField + EmbeddedModelField + Meta.indexes."""

        class Address(models.Model):
            indexed_one = models.CharField(max_length=10)

            class Meta:
                app_label = "schema_"
                indexes = [models.Index(fields=["indexed_one"])]

        class Author(models.Model):
            address = EmbeddedModelField(Address)
            indexed_two = models.CharField(max_length=10)

            class Meta:
                app_label = "schema_"
                indexes = [models.Index(fields=["indexed_two"])]

        class Book(models.Model):
            author = EmbeddedModelField(Author)

            class Meta:
                app_label = "schema_"

        new_field = EmbeddedModelField(Author)
        new_field.set_attributes_from_name("author")
        with connection.schema_editor() as editor:
            # Create the table and add the field.
            editor.create_model(Book)
            editor.add_field(Book, new_field)
            # Embedded indexes are created.
            self.assertEqual(
                self.get_constraints_for_columns(Book, ["author.indexed_two"]),
                ["schema__aut_indexed_b19137_idx"],
            )
            self.assertEqual(
                self.get_constraints_for_columns(
                    Book,
                    ["author.address.indexed_one"],
                ),
                ["schema__add_indexed_b64972_idx"],
            )
            editor.remove_field(Book, new_field)
            # Embedded indexes are removed.
            self.assertEqual(
                self.get_constraints_for_columns(Book, ["author.indexed_two"]),
                [],
            )
            self.assertEqual(
                self.get_constraints_for_columns(
                    Book,
                    ["author.address.indexed_one"],
                ),
                [],
            )
            editor.delete_model(Author)
        self.assertTableNotExists(Author)

    @isolate_apps("schema_")
    def test_add_remove_field_constraints(self):
        """AddField/RemoveField + EmbeddedModelField + Meta.constraints."""

        class Address(models.Model):
            unique_constraint_one = models.CharField(max_length=10)

            class Meta:
                app_label = "schema_"
                constraints = [
                    models.UniqueConstraint(fields=["unique_constraint_one"], name="unique_one")
                ]

        class Author(models.Model):
            address = EmbeddedModelField(Address)
            unique_constraint_two = models.CharField(max_length=10)

            class Meta:
                app_label = "schema_"
                constraints = [
                    models.UniqueConstraint(fields=["unique_constraint_two"], name="unique_two")
                ]

        class Book(models.Model):
            author = EmbeddedModelField(Author)

            class Meta:
                app_label = "schema_"

        new_field = EmbeddedModelField(Author)
        new_field.set_attributes_from_name("author")
        with connection.schema_editor() as editor:
            # Create the table and add the field.
            editor.create_model(Book)
            editor.add_field(Book, new_field)
            # Embedded constraints are created.
            self.assertEqual(
                self.get_constraints_for_columns(Book, ["author.unique_constraint_two"]),
                ["unique_two"],
            )
            self.assertEqual(
                self.get_constraints_for_columns(
                    Book,
                    ["author.address.unique_constraint_one"],
                ),
                ["unique_one"],
            )
            editor.remove_field(Book, new_field)
            # Embedded constraints are removed.
            self.assertEqual(
                self.get_constraints_for_columns(Book, ["author.unique_constraint_two"]),
                [],
            )
            self.assertEqual(
                self.get_constraints_for_columns(
                    Book,
                    ["author.address.unique_constraint_one"],
                ),
                [],
            )
            editor.delete_model(Author)
        self.assertTableNotExists(Author)


class EmbeddedModelsIgnoredTests(TestMixin, TransactionTestCase):
    def test_embedded_not_created(self):
        """create_model() and delete_model() ignore EmbeddedModel."""
        with connection.schema_editor() as editor:
            editor.create_model(Book)
            editor.create_model(Address)
            editor.create_model(Author)
            self.assertTableExists(Book)
            self.assertTableNotExists(Address)
            self.assertTableNotExists(Author)
            editor.delete_model(Book)
            with self.assertNumQueries(0):
                editor.delete_model(Address)
                editor.delete_model(Author)
        self.assertTableNotExists(Book)

    def test_embedded_add_field_ignored(self):
        """add_field() and remove_field() ignore EmbeddedModel."""
        new_field = models.CharField(max_length=1, default="a")
        new_field.set_attributes_from_name("char")
        with connection.schema_editor() as editor, self.assertNumQueries(0):
            editor.add_field(Author, new_field)
        with connection.schema_editor() as editor, self.assertNumQueries(0):
            editor.remove_field(Author, new_field)

    def test_embedded_alter_field_ignored(self):
        """alter_field() ignores EmbeddedModel."""
        old_field = models.CharField(max_length=1)
        old_field.set_attributes_from_name("old")
        new_field = models.CharField(max_length=1)
        new_field.set_attributes_from_name("new")
        with connection.schema_editor() as editor, self.assertNumQueries(0):
            editor.alter_field(Author, old_field, new_field)



"""These tests are forked from Django's tests/cache/tests.py."""
import os
import pickle
import time
from functools import wraps
from unittest import mock

from bson import SON
from django.conf import settings
from django.core import management
from django.core.cache import DEFAULT_CACHE_ALIAS, CacheKeyWarning, cache, caches
from django.core.cache.backends.base import InvalidCacheBackendError
from django.http import HttpResponse
from django.middleware.cache import FetchFromCacheMiddleware, UpdateCacheMiddleware
from django.test import RequestFactory, TestCase, modify_settings, override_settings

from .models import Poll, expensive_calculation

KEY_ERRORS_WITH_MEMCACHED_MSG = (
    "Cache key contains characters that will cause errors if used with memcached: %r"
)


def f():
    return 42


class C:
    def m(n):
        return 24


class Unpicklable:
    def __getstate__(self):
        raise pickle.PickleError()


def empty_response(request):  # noqa: ARG001
    return HttpResponse()


def retry(retries=3, delay=1):
    def decorator(func):
        @wraps(func)
        def wrapper(*args, **kwargs):
            attempts = 0
            while attempts < retries:
                try:
                    return func(*args, **kwargs)
                except AssertionError:
                    attempts += 1
                    if attempts >= retries:
                        raise
                    time.sleep(delay)
            return None

        return wrapper

    return decorator


def custom_key_func(key, key_prefix, version):
    "A customized cache key function"
    return "CUSTOM-" + "-".join([key_prefix, str(version), key])


_caches_setting_base = {
    "default": {},
    "prefix": {"KEY_PREFIX": f"cacheprefix{os.getpid()}"},
    "v2": {"VERSION": 2},
    "custom_key": {"KEY_FUNCTION": custom_key_func},
    "custom_key2": {"KEY_FUNCTION": "cache_.tests.custom_key_func"},
    "cull": {"OPTIONS": {"MAX_ENTRIES": 30}},
    "zero_cull": {"OPTIONS": {"CULL_FREQUENCY": 0, "MAX_ENTRIES": 30}},
}


def caches_setting_for_tests(base=None, exclude=None, **params):
    # `base` is used to pull in the memcached config from the original settings,
    # `exclude` is a set of cache names denoting which `_caches_setting_base` keys
    # should be omitted.
    # `params` are test specific overrides and `_caches_settings_base` is the
    # base config for the tests.
    # This results in the following search order:
    # params -> _caches_setting_base -> base
    base = base or {}
    exclude = exclude or set()
    setting = {k: base.copy() for k in _caches_setting_base if k not in exclude}
    for key, cache_params in setting.items():
        cache_params.update(_caches_setting_base[key])
        cache_params.update(params)
    return setting


@override_settings(
    CACHES=caches_setting_for_tests(
        BACKEND="django_mongodb_backend.cache.MongoDBCache",
        # Spaces are used in the name to ensure quoting/escaping works.
        LOCATION="test cache collection",
    ),
)
@modify_settings(
    INSTALLED_APPS={"prepend": "django_mongodb_backend"},
)
class CacheTests(TestCase):
    factory = RequestFactory()
    incr_decr_type_error_msg = "Cannot apply %s() to a non-numeric value."

    def setUp(self):
        # The super calls needs to happen first for the settings override.
        super().setUp()
        self.create_cache_collection()
        self.addCleanup(self.drop_collection)

    def create_cache_collection(self):
        management.call_command("createcachecollection", verbosity=0)

    def drop_collection(self):
        cache.collection_for_write.drop()

    def test_simple(self):
        # Simple cache set/get works
        cache.set("key", "value")
        self.assertEqual(cache.get("key"), "value")

    def test_default_used_when_none_is_set(self):
        """If None is cached, get() returns it instead of the default."""
        cache.set("key_default_none", None)
        self.assertIsNone(cache.get("key_default_none", default="default"))

    def test_add(self):
        # A key can be added to a cache
        self.assertIs(cache.add("addkey1", "value"), True)
        self.assertIs(cache.add("addkey1", "newvalue"), False)
        self.assertEqual(cache.get("addkey1"), "value")

    def test_prefix(self):
        # Test for same cache key conflicts between shared backend
        cache.set("somekey", "value")

        # should not be set in the prefixed cache
        self.assertIs(caches["prefix"].has_key("somekey"), False)

        caches["prefix"].set("somekey", "value2")

        self.assertEqual(cache.get("somekey"), "value")
        self.assertEqual(caches["prefix"].get("somekey"), "value2")

    def test_non_existent(self):
        """Nonexistent cache keys return as None/default."""
        self.assertIsNone(cache.get("does_not_exist"))
        self.assertEqual(cache.get("does_not_exist", "bang!"), "bang!")

    def test_get_many(self):
        # Multiple cache keys can be returned using get_many
        cache.set_many({"a": "a", "b": "b", "c": "c", "d": "d"})
        self.assertEqual(cache.get_many(["a", "c", "d"]), {"a": "a", "c": "c", "d": "d"})
        self.assertEqual(cache.get_many(["a", "b", "e"]), {"a": "a", "b": "b"})
        self.assertEqual(cache.get_many(iter(["a", "b", "e"])), {"a": "a", "b": "b"})
        cache.set_many({"x": None, "y": 1})
        self.assertEqual(cache.get_many(["x", "y"]), {"x": None, "y": 1})

    def test_delete(self):
        # Cache keys can be deleted
        cache.set_many({"key1": "spam", "key2": "eggs"})
        self.assertEqual(cache.get("key1"), "spam")
        self.assertIs(cache.delete("key1"), True)
        self.assertIsNone(cache.get("key1"))
        self.assertEqual(cache.get("key2"), "eggs")

    def test_delete_nonexistent(self):
        self.assertIs(cache.delete("nonexistent_key"), False)

    def test_has_key(self):
        # The cache can be inspected for cache keys
        cache.set("hello1", "goodbye1")
        self.assertIs(cache.has_key("hello1"), True)
        self.assertIs(cache.has_key("goodbye1"), False)
        cache.set("no_expiry", "here", None)
        self.assertIs(cache.has_key("no_expiry"), True)
        cache.set("null", None)
        self.assertIs(cache.has_key("null"), True)

    def test_in(self):
        # The in operator can be used to inspect cache contents
        cache.set("hello2", "goodbye2")
        self.assertIn("hello2", cache)
        self.assertNotIn("goodbye2", cache)
        cache.set("null", None)
        self.assertIn("null", cache)

    def test_incr(self):
        # Cache values can be incremented
        cache.set("answer", 41)
        self.assertEqual(cache.incr("answer"), 42)
        self.assertEqual(cache.get("answer"), 42)
        self.assertEqual(cache.incr("answer", 10), 52)
        self.assertEqual(cache.get("answer"), 52)
        self.assertEqual(cache.incr("answer", -10), 42)
        with self.assertRaisesMessage(ValueError, "Key 'does_not_exist' not found."):
            cache.incr("does_not_exist")
        with self.assertRaisesMessage(ValueError, "Key 'does_not_exist' not found."):
            cache.incr("does_not_exist", -1)
        cache.set("null", None)
        with self.assertRaisesMessage(TypeError, self.incr_decr_type_error_msg % "incr"):
            cache.incr("null")

    def test_decr(self):
        # Cache values can be decremented
        cache.set("answer", 43)
        self.assertEqual(cache.decr("answer"), 42)
        self.assertEqual(cache.get("answer"), 42)
        self.assertEqual(cache.decr("answer", 10), 32)
        self.assertEqual(cache.get("answer"), 32)
        self.assertEqual(cache.decr("answer", -10), 42)
        with self.assertRaisesMessage(ValueError, "Key 'does_not_exist' not found."):
            cache.decr("does_not_exist")
        with self.assertRaisesMessage(ValueError, "Key 'does_not_exist' not found."):
            cache.incr("does_not_exist", -1)
        cache.set("null", None)
        with self.assertRaisesMessage(TypeError, self.incr_decr_type_error_msg % "decr"):
            cache.decr("null")

    def test_close(self):
        self.assertTrue(hasattr(cache, "close"))
        cache.close()

    def test_data_types(self):
        # Many different data types can be cached
        tests = {
            "string": "this is a string",
            "int": 42,
            "bool": True,
            "list": [1, 2, 3, 4],
            "tuple": (1, 2, 3, 4),
            "dict": {"A": 1, "B": 2},
            "function": f,
            "class": C,
        }
        for key, value in tests.items():
            with self.subTest(key=key):
                cache.set(key, value)
                self.assertEqual(cache.get(key), value)

    def test_cache_read_for_model_instance(self):
        # Don't want fields with callable as default to be called on cache read
        expensive_calculation.num_runs = 0
        Poll.objects.all().delete()
        my_poll = Poll.objects.create(question="Well?")
        self.assertEqual(Poll.objects.count(), 1)
        pub_date = my_poll.pub_date
        cache.set("question", my_poll)
        cached_poll = cache.get("question")
        self.assertEqual(cached_poll.pub_date, pub_date)
        # We only want the default expensive calculation run once
        self.assertEqual(expensive_calculation.num_runs, 1)

    def test_cache_write_for_model_instance_with_deferred(self):
        # Don't want fields with callable as default to be called on cache write
        expensive_calculation.num_runs = 0
        Poll.objects.all().delete()
        Poll.objects.create(question="What?")
        self.assertEqual(expensive_calculation.num_runs, 1)
        defer_qs = Poll.objects.defer("question")
        self.assertEqual(defer_qs.count(), 1)
        self.assertEqual(expensive_calculation.num_runs, 1)
        cache.set("deferred_queryset", defer_qs)
        # cache set should not re-evaluate default functions
        self.assertEqual(expensive_calculation.num_runs, 1)

    def test_cache_read_for_model_instance_with_deferred(self):
        # Don't want fields with callable as default to be called on cache read
        expensive_calculation.num_runs = 0
        Poll.objects.all().delete()
        Poll.objects.create(question="What?")
        self.assertEqual(expensive_calculation.num_runs, 1)
        defer_qs = Poll.objects.defer("question")
        self.assertEqual(defer_qs.count(), 1)
        cache.set("deferred_queryset", defer_qs)
        self.assertEqual(expensive_calculation.num_runs, 1)
        runs_before_cache_read = expensive_calculation.num_runs
        cache.get("deferred_queryset")
        # We only want the default expensive calculation run on creation and set
        self.assertEqual(expensive_calculation.num_runs, runs_before_cache_read)

    def test_expiration(self):
        # Cache values can be set to expire
        cache.set("expire1", "very quickly", 1)
        cache.set("expire2", "very quickly", 1)
        cache.set("expire3", "very quickly", 1)

        time.sleep(2)
        self.assertIsNone(cache.get("expire1"))

        self.assertIs(cache.add("expire2", "newvalue"), True)
        self.assertEqual(cache.get("expire2"), "newvalue")
        self.assertIs(cache.has_key("expire3"), False)

    @retry()
    def test_touch(self):
        # cache.touch() updates the timeout.
        cache.set("expire1", "very quickly", timeout=1)
        self.assertIs(cache.touch("expire1", timeout=4), True)
        time.sleep(2)
        self.assertIs(cache.has_key("expire1"), True)
        time.sleep(3)
        self.assertIs(cache.has_key("expire1"), False)
        # cache.touch() works without the timeout argument.
        cache.set("expire1", "very quickly", timeout=1)
        self.assertIs(cache.touch("expire1"), True)
        time.sleep(2)
        self.assertIs(cache.has_key("expire1"), True)

        self.assertIs(cache.touch("nonexistent"), False)

    def test_unicode(self):
        # Unicode values can be cached
        stuff = {
            "ascii": "ascii_value",
            "unicode_ascii": "Iñtërnâtiônàlizætiøn1",
            "Iñtërnâtiônàlizætiøn": "Iñtërnâtiônàlizætiøn2",
            "ascii2": {"x": 1},
        }
        # Test `set`
        for key, value in stuff.items():
            with self.subTest(key=key):
                cache.set(key, value)
                self.assertEqual(cache.get(key), value)

        # Test `add`
        for key, value in stuff.items():
            with self.subTest(key=key):
                self.assertIs(cache.delete(key), True)
                self.assertIs(cache.add(key, value), True)
                self.assertEqual(cache.get(key), value)

        # Test `set_many`
        for key in stuff:
            self.assertIs(cache.delete(key), True)
        cache.set_many(stuff)
        for key, value in stuff.items():
            with self.subTest(key=key):
                self.assertEqual(cache.get(key), value)

    def test_binary_string(self):
        # Binary strings should be cacheable
        from zlib import compress, decompress

        value = "value_to_be_compressed"
        compressed_value = compress(value.encode())

        # Test set
        cache.set("binary1", compressed_value)
        compressed_result = cache.get("binary1")
        self.assertEqual(compressed_value, compressed_result)
        self.assertEqual(value, decompress(compressed_result).decode())

        # Test add
        self.assertIs(cache.add("binary1-add", compressed_value), True)
        compressed_result = cache.get("binary1-add")
        self.assertEqual(compressed_value, compressed_result)
        self.assertEqual(value, decompress(compressed_result).decode())

        # Test set_many
        cache.set_many({"binary1-set_many": compressed_value})
        compressed_result = cache.get("binary1-set_many")
        self.assertEqual(compressed_value, compressed_result)
        self.assertEqual(value, decompress(compressed_result).decode())

    def test_set_many(self):
        # Multiple keys can be set using set_many
        cache.set_many({"key1": "spam", "key2": "eggs"})
        self.assertEqual(cache.get("key1"), "spam")
        self.assertEqual(cache.get("key2"), "eggs")

    def test_set_many_returns_empty_list_on_success(self):
        """set_many() returns an empty list when all keys are inserted."""
        failing_keys = cache.set_many({"key1": "spam", "key2": "eggs"})
        self.assertEqual(failing_keys, [])

    def test_set_many_expiration(self):
        # set_many takes a second ``timeout`` parameter
        cache.set_many({"key1": "spam", "key2": "eggs"}, 1)
        time.sleep(2)
        self.assertIsNone(cache.get("key1"))
        self.assertIsNone(cache.get("key2"))

    def test_set_many_empty_data(self):
        self.assertEqual(cache.set_many({}), [])

    def test_delete_many(self):
        # Multiple keys can be deleted using delete_many
        cache.set_many({"key1": "spam", "key2": "eggs", "key3": "ham"})
        cache.delete_many(["key1", "key2"])
        self.assertIsNone(cache.get("key1"))
        self.assertIsNone(cache.get("key2"))
        self.assertEqual(cache.get("key3"), "ham")

    def test_delete_many_no_keys(self):
        self.assertIsNone(cache.delete_many([]))

    def test_clear(self):
        # The cache can be emptied using clear
        cache.set_many({"key1": "spam", "key2": "eggs"})
        cache.clear()
        self.assertIsNone(cache.get("key1"))
        self.assertIsNone(cache.get("key2"))

    def test_long_timeout(self):
        """
        Follow memcached's convention where a timeout greater than 30 days is
        treated as an absolute expiration timestamp instead of a relative
        offset (#12399).
        """
        cache.set("key1", "eggs", 60 * 60 * 24 * 30 + 1)  # 30 days + 1 second
        self.assertEqual(cache.get("key1"), "eggs")

        self.assertIs(cache.add("key2", "ham", 60 * 60 * 24 * 30 + 1), True)
        self.assertEqual(cache.get("key2"), "ham")

        cache.set_many({"key3": "sausage", "key4": "lobster bisque"}, 60 * 60 * 24 * 30 + 1)
        self.assertEqual(cache.get("key3"), "sausage")
        self.assertEqual(cache.get("key4"), "lobster bisque")

    @retry()
    def test_forever_timeout(self):
        """
        Passing in None into timeout results in a value that is cached forever
        """
        cache.set("key1", "eggs", None)
        self.assertEqual(cache.get("key1"), "eggs")

        self.assertIs(cache.add("key2", "ham", None), True)
        self.assertEqual(cache.get("key2"), "ham")
        self.assertIs(cache.add("key1", "new eggs", None), False)
        self.assertEqual(cache.get("key1"), "eggs")

        cache.set_many({"key3": "sausage", "key4": "lobster bisque"}, None)
        self.assertEqual(cache.get("key3"), "sausage")
        self.assertEqual(cache.get("key4"), "lobster bisque")

        cache.set("key5", "belgian fries", timeout=1)
        self.assertIs(cache.touch("key5", timeout=None), True)
        time.sleep(2)
        self.assertEqual(cache.get("key5"), "belgian fries")

    def test_zero_timeout(self):
        """
        Passing in zero into timeout results in a value that is not cached
        """
        cache.set("key1", "eggs", 0)
        self.assertIsNone(cache.get("key1"))

        self.assertIs(cache.add("key2", "ham", 0), True)
        self.assertIsNone(cache.get("key2"))

        cache.set_many({"key3": "sausage", "key4": "lobster bisque"}, 0)
        self.assertIsNone(cache.get("key3"))
        self.assertIsNone(cache.get("key4"))

        cache.set("key5", "belgian fries", timeout=5)
        self.assertIs(cache.touch("key5", timeout=0), True)
        self.assertIsNone(cache.get("key5"))

    def test_float_timeout(self):
        # Make sure a timeout given as a float doesn't crash anything.
        cache.set("key1", "spam", 100.2)
        self.assertEqual(cache.get("key1"), "spam")

    def _perform_cull_test(self, cull_cache_name, initial_count, final_count):
        try:
            cull_cache = caches[cull_cache_name]
        except InvalidCacheBackendError:
            self.skipTest("Culling isn't implemented.")

        # Create initial cache key entries. This will overflow the cache,
        # causing a cull.
        for i in range(1, initial_count):
            cull_cache.set("cull%d" % i, "value", 1000)
        count = 0
        # Count how many keys are left in the cache.
        for i in range(1, initial_count):
            if cull_cache.has_key("cull%d" % i):
                count += 1
        self.assertEqual(count, final_count)

    def test_cull(self):
        self._perform_cull_test("cull", 50, 29)

    def test_zero_cull(self):
        self._perform_cull_test("zero_cull", 50, 19)

    def test_cull_delete_when_store_empty(self):
        try:
            cull_cache = caches["cull"]
        except InvalidCacheBackendError:
            self.skipTest("Culling isn't implemented.")
        old_max_entries = cull_cache._max_entries
        # Force _cull to delete on first cached record.
        cull_cache._max_entries = -1
        try:
            cull_cache.set("force_cull_delete", "value", 1000)
            self.assertIs(cull_cache.has_key("force_cull_delete"), True)
        finally:
            cull_cache._max_entries = old_max_entries

    def _perform_invalid_key_test(self, key, expected_warning, key_func=None):
        """
        All the builtin backends should warn (except memcached that should
        error) on keys that would be refused by memcached. This encourages
        portable caching code without making it too difficult to use production
        backends with more liberal key rules. Refs #6447.
        """

        # mimic custom ``make_key`` method being defined since the default will
        # never show the below warnings
        def func(key, *args):  # noqa: ARG001
            return key

        old_func = cache.key_func
        cache.key_func = key_func or func

        tests = [
            ("add", [key, 1]),
            ("get", [key]),
            ("set", [key, 1]),
            ("incr", [key]),
            ("decr", [key]),
            ("touch", [key]),
            ("delete", [key]),
            ("get_many", [[key, "b"]]),
            ("set_many", [{key: 1, "b": 2}]),
            ("delete_many", [[key, "b"]]),
        ]
        try:
            for operation, args in tests:
                with self.subTest(operation=operation):
                    with self.assertWarns(CacheKeyWarning) as cm:
                        getattr(cache, operation)(*args)
                    self.assertEqual(str(cm.warning), expected_warning)
        finally:
            cache.key_func = old_func

    def test_invalid_key_characters(self):
        # memcached doesn't allow whitespace or control characters in keys.
        key = "key with spaces and 清"
        self._perform_invalid_key_test(key, KEY_ERRORS_WITH_MEMCACHED_MSG % key)

    def test_invalid_key_length(self):
        # memcached limits key length to 250.
        key = ("a" * 250) + "清"
        expected_warning = (
            "Cache key will cause errors if used with memcached: " f"'{key}' (longer than 250)"
        )
        self._perform_invalid_key_test(key, expected_warning)

    def test_invalid_with_version_key_length(self):
        # Custom make_key() that adds a version to the key and exceeds the
        # limit.
        def key_func(key, *args):  # noqa: ARG001
            return key + ":1"

        key = "a" * 249
        expected_warning = (
            "Cache key will cause errors if used with memcached: "
            f"'{key_func(key)}' (longer than 250)"
        )
        self._perform_invalid_key_test(key, expected_warning, key_func=key_func)

    def test_cache_versioning_get_set(self):
        # set, using default version = 1
        cache.set("answer1", 42)
        self.assertEqual(cache.get("answer1"), 42)
        self.assertEqual(cache.get("answer1", version=1), 42)
        self.assertIsNone(cache.get("answer1", version=2))

        self.assertIsNone(caches["v2"].get("answer1"))
        self.assertEqual(caches["v2"].get("answer1", version=1), 42)
        self.assertIsNone(caches["v2"].get("answer1", version=2))

        # set, default version = 1, but manually override version = 2
        cache.set("answer2", 42, version=2)
        self.assertIsNone(cache.get("answer2"))
        self.assertIsNone(cache.get("answer2", version=1))
        self.assertEqual(cache.get("answer2", version=2), 42)

        self.assertEqual(caches["v2"].get("answer2"), 42)
        self.assertIsNone(caches["v2"].get("answer2", version=1))
        self.assertEqual(caches["v2"].get("answer2", version=2), 42)

        # v2 set, using default version = 2
        caches["v2"].set("answer3", 42)
        self.assertIsNone(cache.get("answer3"))
        self.assertIsNone(cache.get("answer3", version=1))
        self.assertEqual(cache.get("answer3", version=2), 42)

        self.assertEqual(caches["v2"].get("answer3"), 42)
        self.assertIsNone(caches["v2"].get("answer3", version=1))
        self.assertEqual(caches["v2"].get("answer3", version=2), 42)

        # v2 set, default version = 2, but manually override version = 1
        caches["v2"].set("answer4", 42, version=1)
        self.assertEqual(cache.get("answer4"), 42)
        self.assertEqual(cache.get("answer4", version=1), 42)
        self.assertIsNone(cache.get("answer4", version=2))

        self.assertIsNone(caches["v2"].get("answer4"))
        self.assertEqual(caches["v2"].get("answer4", version=1), 42)
        self.assertIsNone(caches["v2"].get("answer4", version=2))

    def test_cache_versioning_add(self):
        # add, default version = 1, but manually override version = 2
        self.assertIs(cache.add("answer1", 42, version=2), True)
        self.assertIsNone(cache.get("answer1", version=1))
        self.assertEqual(cache.get("answer1", version=2), 42)

        self.assertIs(cache.add("answer1", 37, version=2), False)
        self.assertIsNone(cache.get("answer1", version=1))
        self.assertEqual(cache.get("answer1", version=2), 42)

        self.assertIs(cache.add("answer1", 37, version=1), True)
        self.assertEqual(cache.get("answer1", version=1), 37)
        self.assertEqual(cache.get("answer1", version=2), 42)

        # v2 add, using default version = 2
        self.assertIs(caches["v2"].add("answer2", 42), True)
        self.assertIsNone(cache.get("answer2", version=1))
        self.assertEqual(cache.get("answer2", version=2), 42)

        self.assertIs(caches["v2"].add("answer2", 37), False)
        self.assertIsNone(cache.get("answer2", version=1))
        self.assertEqual(cache.get("answer2", version=2), 42)

        self.assertIs(caches["v2"].add("answer2", 37, version=1), True)
        self.assertEqual(cache.get("answer2", version=1), 37)
        self.assertEqual(cache.get("answer2", version=2), 42)

        # v2 add, default version = 2, but manually override version = 1
        self.assertIs(caches["v2"].add("answer3", 42, version=1), True)
        self.assertEqual(cache.get("answer3", version=1), 42)
        self.assertIsNone(cache.get("answer3", version=2))

        self.assertIs(caches["v2"].add("answer3", 37, version=1), False)
        self.assertEqual(cache.get("answer3", version=1), 42)
        self.assertIsNone(cache.get("answer3", version=2))

        self.assertIs(caches["v2"].add("answer3", 37), True)
        self.assertEqual(cache.get("answer3", version=1), 42)
        self.assertEqual(cache.get("answer3", version=2), 37)

    def test_cache_versioning_has_key(self):
        cache.set("answer1", 42)

        # has_key
        self.assertIs(cache.has_key("answer1"), True)
        self.assertIs(cache.has_key("answer1", version=1), True)
        self.assertIs(cache.has_key("answer1", version=2), False)

        self.assertIs(caches["v2"].has_key("answer1"), False)
        self.assertIs(caches["v2"].has_key("answer1", version=1), True)
        self.assertIs(caches["v2"].has_key("answer1", version=2), False)

    def test_cache_versioning_delete(self):
        cache.set("answer1", 37, version=1)
        cache.set("answer1", 42, version=2)
        self.assertIs(cache.delete("answer1"), True)
        self.assertIsNone(cache.get("answer1", version=1))
        self.assertEqual(cache.get("answer1", version=2), 42)

        cache.set("answer2", 37, version=1)
        cache.set("answer2", 42, version=2)
        self.assertIs(cache.delete("answer2", version=2), True)
        self.assertEqual(cache.get("answer2", version=1), 37)
        self.assertIsNone(cache.get("answer2", version=2))

        cache.set("answer3", 37, version=1)
        cache.set("answer3", 42, version=2)
        self.assertIs(caches["v2"].delete("answer3"), True)
        self.assertEqual(cache.get("answer3", version=1), 37)
        self.assertIsNone(cache.get("answer3", version=2))

        cache.set("answer4", 37, version=1)
        cache.set("answer4", 42, version=2)
        self.assertIs(caches["v2"].delete("answer4", version=1), True)
        self.assertIsNone(cache.get("answer4", version=1))
        self.assertEqual(cache.get("answer4", version=2), 42)

    def test_cache_versioning_incr_decr(self):
        cache.set("answer1", 37, version=1)
        cache.set("answer1", 42, version=2)
        self.assertEqual(cache.incr("answer1"), 38)
        self.assertEqual(cache.get("answer1", version=1), 38)
        self.assertEqual(cache.get("answer1", version=2), 42)
        self.assertEqual(cache.decr("answer1"), 37)
        self.assertEqual(cache.get("answer1", version=1), 37)
        self.assertEqual(cache.get("answer1", version=2), 42)

        cache.set("answer2", 37, version=1)
        cache.set("answer2", 42, version=2)
        self.assertEqual(cache.incr("answer2", version=2), 43)
        self.assertEqual(cache.get("answer2", version=1), 37)
        self.assertEqual(cache.get("answer2", version=2), 43)
        self.assertEqual(cache.decr("answer2", version=2), 42)
        self.assertEqual(cache.get("answer2", version=1), 37)
        self.assertEqual(cache.get("answer2", version=2), 42)

        cache.set("answer3", 37, version=1)
        cache.set("answer3", 42, version=2)
        self.assertEqual(caches["v2"].incr("answer3"), 43)
        self.assertEqual(cache.get("answer3", version=1), 37)
        self.assertEqual(cache.get("answer3", version=2), 43)
        self.assertEqual(caches["v2"].decr("answer3"), 42)
        self.assertEqual(cache.get("answer3", version=1), 37)
        self.assertEqual(cache.get("answer3", version=2), 42)

        cache.set("answer4", 37, version=1)
        cache.set("answer4", 42, version=2)
        self.assertEqual(caches["v2"].incr("answer4", version=1), 38)
        self.assertEqual(cache.get("answer4", version=1), 38)
        self.assertEqual(cache.get("answer4", version=2), 42)
        self.assertEqual(caches["v2"].decr("answer4", version=1), 37)
        self.assertEqual(cache.get("answer4", version=1), 37)
        self.assertEqual(cache.get("answer4", version=2), 42)

    def test_cache_versioning_get_set_many(self):
        # set, using default version = 1
        cache.set_many({"ford1": 37, "arthur1": 42})
        self.assertEqual(cache.get_many(["ford1", "arthur1"]), {"ford1": 37, "arthur1": 42})
        self.assertEqual(
            cache.get_many(["ford1", "arthur1"], version=1),
            {"ford1": 37, "arthur1": 42},
        )
        self.assertEqual(cache.get_many(["ford1", "arthur1"], version=2), {})

        self.assertEqual(caches["v2"].get_many(["ford1", "arthur1"]), {})
        self.assertEqual(
            caches["v2"].get_many(["ford1", "arthur1"], version=1),
            {"ford1": 37, "arthur1": 42},
        )
        self.assertEqual(caches["v2"].get_many(["ford1", "arthur1"], version=2), {})

        # set, default version = 1, but manually override version = 2
        cache.set_many({"ford2": 37, "arthur2": 42}, version=2)
        self.assertEqual(cache.get_many(["ford2", "arthur2"]), {})
        self.assertEqual(cache.get_many(["ford2", "arthur2"], version=1), {})
        self.assertEqual(
            cache.get_many(["ford2", "arthur2"], version=2),
            {"ford2": 37, "arthur2": 42},
        )

        self.assertEqual(caches["v2"].get_many(["ford2", "arthur2"]), {"ford2": 37, "arthur2": 42})
        self.assertEqual(caches["v2"].get_many(["ford2", "arthur2"], version=1), {})
        self.assertEqual(
            caches["v2"].get_many(["ford2", "arthur2"], version=2),
            {"ford2": 37, "arthur2": 42},
        )

        # v2 set, using default version = 2
        caches["v2"].set_many({"ford3": 37, "arthur3": 42})
        self.assertEqual(cache.get_many(["ford3", "arthur3"]), {})
        self.assertEqual(cache.get_many(["ford3", "arthur3"], version=1), {})
        self.assertEqual(
            cache.get_many(["ford3", "arthur3"], version=2),
            {"ford3": 37, "arthur3": 42},
        )

        self.assertEqual(caches["v2"].get_many(["ford3", "arthur3"]), {"ford3": 37, "arthur3": 42})
        self.assertEqual(caches["v2"].get_many(["ford3", "arthur3"], version=1), {})
        self.assertEqual(
            caches["v2"].get_many(["ford3", "arthur3"], version=2),
            {"ford3": 37, "arthur3": 42},
        )

        # v2 set, default version = 2, but manually override version = 1
        caches["v2"].set_many({"ford4": 37, "arthur4": 42}, version=1)
        self.assertEqual(cache.get_many(["ford4", "arthur4"]), {"ford4": 37, "arthur4": 42})
        self.assertEqual(
            cache.get_many(["ford4", "arthur4"], version=1),
            {"ford4": 37, "arthur4": 42},
        )
        self.assertEqual(cache.get_many(["ford4", "arthur4"], version=2), {})

        self.assertEqual(caches["v2"].get_many(["ford4", "arthur4"]), {})
        self.assertEqual(
            caches["v2"].get_many(["ford4", "arthur4"], version=1),
            {"ford4": 37, "arthur4": 42},
        )
        self.assertEqual(caches["v2"].get_many(["ford4", "arthur4"], version=2), {})

    def test_incr_version(self):
        cache.set("answer", 42, version=2)
        self.assertIsNone(cache.get("answer"))
        self.assertIsNone(cache.get("answer", version=1))
        self.assertEqual(cache.get("answer", version=2), 42)
        self.assertIsNone(cache.get("answer", version=3))

        self.assertEqual(cache.incr_version("answer", version=2), 3)
        self.assertIsNone(cache.get("answer"))
        self.assertIsNone(cache.get("answer", version=1))
        self.assertIsNone(cache.get("answer", version=2))
        self.assertEqual(cache.get("answer", version=3), 42)

        caches["v2"].set("answer2", 42)
        self.assertEqual(caches["v2"].get("answer2"), 42)
        self.assertIsNone(caches["v2"].get("answer2", version=1))
        self.assertEqual(caches["v2"].get("answer2", version=2), 42)
        self.assertIsNone(caches["v2"].get("answer2", version=3))

        self.assertEqual(caches["v2"].incr_version("answer2"), 3)
        self.assertIsNone(caches["v2"].get("answer2"))
        self.assertIsNone(caches["v2"].get("answer2", version=1))
        self.assertIsNone(caches["v2"].get("answer2", version=2))
        self.assertEqual(caches["v2"].get("answer2", version=3), 42)

        with self.assertRaises(ValueError):
            cache.incr_version("does_not_exist")

        cache.set("null", None)
        self.assertEqual(cache.incr_version("null"), 2)

    def test_decr_version(self):
        cache.set("answer", 42, version=2)
        self.assertIsNone(cache.get("answer"))
        self.assertIsNone(cache.get("answer", version=1))
        self.assertEqual(cache.get("answer", version=2), 42)

        self.assertEqual(cache.decr_version("answer", version=2), 1)
        self.assertEqual(cache.get("answer"), 42)
        self.assertEqual(cache.get("answer", version=1), 42)
        self.assertIsNone(cache.get("answer", version=2))

        caches["v2"].set("answer2", 42)
        self.assertEqual(caches["v2"].get("answer2"), 42)
        self.assertIsNone(caches["v2"].get("answer2", version=1))
        self.assertEqual(caches["v2"].get("answer2", version=2), 42)

        self.assertEqual(caches["v2"].decr_version("answer2"), 1)
        self.assertIsNone(caches["v2"].get("answer2"))
        self.assertEqual(caches["v2"].get("answer2", version=1), 42)
        self.assertIsNone(caches["v2"].get("answer2", version=2))

        with self.assertRaises(ValueError):
            cache.decr_version("does_not_exist", version=2)

        cache.set("null", None, version=2)
        self.assertEqual(cache.decr_version("null", version=2), 1)

    def test_custom_key_func(self):
        # Two caches with different key functions aren't visible to each other
        cache.set("answer1", 42)
        self.assertEqual(cache.get("answer1"), 42)
        self.assertIsNone(caches["custom_key"].get("answer1"))
        self.assertIsNone(caches["custom_key2"].get("answer1"))

        caches["custom_key"].set("answer2", 42)
        self.assertIsNone(cache.get("answer2"))
        self.assertEqual(caches["custom_key"].get("answer2"), 42)
        self.assertEqual(caches["custom_key2"].get("answer2"), 42)

    @override_settings(CACHE_MIDDLEWARE_ALIAS=DEFAULT_CACHE_ALIAS)
    def test_cache_write_unpicklable_object(self):
        fetch_middleware = FetchFromCacheMiddleware(empty_response)

        request = self.factory.get("/cache/test")
        request._cache_update_cache = True
        get_cache_data = FetchFromCacheMiddleware(empty_response).process_request(request)
        self.assertIsNone(get_cache_data)

        content = "Testing cookie serialization."

        def get_response(req):  # noqa: ARG001
            response = HttpResponse(content)
            response.set_cookie("foo", "bar")
            return response

        update_middleware = UpdateCacheMiddleware(get_response)
        response = update_middleware(request)

        get_cache_data = fetch_middleware.process_request(request)
        self.assertIsNotNone(get_cache_data)
        self.assertEqual(get_cache_data.content, content.encode())
        self.assertEqual(get_cache_data.cookies, response.cookies)

        UpdateCacheMiddleware(lambda req: get_cache_data)(request)  # noqa: ARG005
        get_cache_data = fetch_middleware.process_request(request)
        self.assertIsNotNone(get_cache_data)
        self.assertEqual(get_cache_data.content, content.encode())
        self.assertEqual(get_cache_data.cookies, response.cookies)

    def test_add_fail_on_pickleerror(self):
        # Shouldn't fail silently if trying to cache an unpicklable type.
        with self.assertRaises(pickle.PickleError):
            cache.add("unpicklable", Unpicklable())

    def test_set_fail_on_pickleerror(self):
        with self.assertRaises(pickle.PickleError):
            cache.set("unpicklable", Unpicklable())

    def test_get_or_set(self):
        self.assertIsNone(cache.get("projector"))
        self.assertEqual(cache.get_or_set("projector", 42), 42)
        self.assertEqual(cache.get("projector"), 42)
        self.assertIsNone(cache.get_or_set("null", None))
        # Previous get_or_set() stores None in the cache.
        self.assertIsNone(cache.get("null", "default"))

    def test_get_or_set_callable(self):
        def my_callable():
            return "value"

        self.assertEqual(cache.get_or_set("mykey", my_callable), "value")
        self.assertEqual(cache.get_or_set("mykey", my_callable()), "value")

        self.assertIsNone(cache.get_or_set("null", lambda: None))
        # Previous get_or_set() stores None in the cache.
        self.assertIsNone(cache.get("null", "default"))

    def test_get_or_set_version(self):
        msg = "get_or_set() missing 1 required positional argument: 'default'"
        self.assertEqual(cache.get_or_set("brian", 1979, version=2), 1979)
        with self.assertRaisesMessage(TypeError, msg):
            cache.get_or_set("brian")
        with self.assertRaisesMessage(TypeError, msg):
            cache.get_or_set("brian", version=1)
        self.assertIsNone(cache.get("brian", version=1))
        self.assertEqual(cache.get_or_set("brian", 42, version=1), 42)
        self.assertEqual(cache.get_or_set("brian", 1979, version=2), 1979)
        self.assertIsNone(cache.get("brian", version=3))

    def test_get_or_set_racing(self):
        with mock.patch(f"{settings.CACHES['default']['BACKEND']}.add") as cache_add:
            # Simulate cache.add() failing to add a value. In that case, the
            # default value should be returned.
            cache_add.return_value = False
            self.assertEqual(cache.get_or_set("key", "default"), "default")

    def test_collection_has_indexes(self):
        indexes = list(cache.collection_for_read.list_indexes())
        self.assertTrue(
            any(
                index["key"] == SON([("expires_at", 1)]) and index.get("expireAfterSeconds") == 0
                for index in indexes
            )
        )
        self.assertTrue(
            any(
                index["key"] == SON([("key", 1)]) and index.get("unique") is True
                for index in indexes
            )
        )

    def test_serializer_dumps(self):
        self.assertEqual(cache.serializer.dumps(123), 123)
        self.assertIsInstance(cache.serializer.dumps(True), bytes)
        self.assertIsInstance(cache.serializer.dumps("abc"), bytes)


class DBCacheRouter:
    """A router that puts the cache table on the 'other' database."""

    def db_for_read(self, model, **hints):
        if model._meta.app_label == "django_cache":
            return "other"
        return None

    def db_for_write(self, model, **hints):
        if model._meta.app_label == "django_cache":
            return "other"
        return None

    def allow_migrate(self, db, app_label, **hints):
        if app_label == "django_cache":
            return db == "other"
        return None


@override_settings(
    CACHES={
        "default": {
            "BACKEND": "django_mongodb_backend.cache.MongoDBCache",
            "LOCATION": "my_cache_table",
        },
    },
)
@modify_settings(
    INSTALLED_APPS={"prepend": "django_mongodb_backend"},
)
class CreateCacheCollectionTests(TestCase):
    databases = {"default", "other"}

    @override_settings(DATABASE_ROUTERS=[DBCacheRouter()])
    def test_createcachetable_observes_database_router(self):
        # cache table should not be created on 'default'
        with self.assertNumQueries(0, using="default"):
            management.call_command("createcachecollection", database="default", verbosity=0)
        # cache table should be created on 'other'
        # Queries:
        #   1: Create indexes
        with self.assertNumQueries(1, using="other"):
            management.call_command("createcachecollection", database="other", verbosity=0)



import json

from bson import ObjectId
from django.core import serializers
from django.core.exceptions import ValidationError
from django.test import SimpleTestCase, TestCase

from django_mongodb_backend import forms
from django_mongodb_backend.fields import ObjectIdField

from .models import NullableObjectIdModel, ObjectIdModel, PrimaryKeyObjectIdModel


class MethodTests(SimpleTestCase):
    def test_deconstruct(self):
        field = ObjectIdField()
        name, path, args, kwargs = field.deconstruct()
        self.assertEqual(path, "django_mongodb_backend.fields.ObjectIdField")
        self.assertEqual(args, [])
        self.assertEqual(kwargs, {})

    def test_formfield(self):
        f = ObjectIdField().formfield()
        self.assertIsInstance(f, forms.ObjectIdField)

    def test_get_internal_type(self):
        f = ObjectIdField()
        self.assertEqual(f.get_internal_type(), "ObjectIdField")

    def test_to_python_string(self):
        value = "1" * 24
        self.assertEqual(ObjectIdField().to_python(value), ObjectId(value))

    def test_to_python_objectid(self):
        value = ObjectId("1" * 24)
        self.assertEqual(ObjectIdField().to_python(value), value)

    def test_to_python_null(self):
        self.assertIsNone(ObjectIdField().to_python(None))

    def test_to_python_invalid_value(self):
        f = ObjectIdField()
        for invalid_value in ["None", "", {}, [], 123]:
            with self.subTest(invalid_value=invalid_value):
                msg = f"['“{invalid_value}” is not a valid Object Id.']"
                with self.assertRaisesMessage(ValidationError, msg):
                    f.to_python(invalid_value)

    def test_get_prep_value_string(self):
        value = "1" * 24
        self.assertEqual(ObjectIdField().get_prep_value(value), ObjectId(value))

    def test_get_prep_value_objectid(self):
        value = ObjectId("1" * 24)
        self.assertEqual(ObjectIdField().get_prep_value(value), value)

    def test_get_prep_value_null(self):
        self.assertIsNone(ObjectIdField().get_prep_value(None))

    def test_get_prep_value_invalid_values(self):
        f = ObjectIdField()
        f.name = "test"
        for invalid_value in ["None", "", {}, [], 123]:
            with self.subTest(invalid_value=invalid_value):
                msg = f"['“{invalid_value}” is not a valid Object Id.']"
                with self.assertRaisesMessage(ValidationError, msg):
                    f.get_prep_value(invalid_value)


class SaveLoadTests(TestCase):
    def test_objectid_instance(self):
        instance = ObjectIdModel.objects.create(field=ObjectId())
        loaded = ObjectIdModel.objects.get()
        self.assertEqual(loaded.field, instance.field)

    def test_str_instance(self):
        ObjectIdModel.objects.create(field="6754ed8e584bc9ceaae3c072")
        loaded = ObjectIdModel.objects.get()
        self.assertEqual(loaded.field, ObjectId("6754ed8e584bc9ceaae3c072"))

    def test_null_handling(self):
        NullableObjectIdModel.objects.create(field=None)
        loaded = NullableObjectIdModel.objects.get()
        self.assertIsNone(loaded.field)

    def test_pk_validated(self):
        with self.assertRaisesMessage(ValidationError, "is not a valid Object Id."):
            PrimaryKeyObjectIdModel.objects.get(pk={})

        with self.assertRaisesMessage(ValidationError, "is not a valid Object Id."):
            PrimaryKeyObjectIdModel.objects.get(pk=[])

    def test_wrong_lookup_type(self):
        with self.assertRaisesMessage(ValidationError, "is not a valid Object Id."):
            ObjectIdModel.objects.get(field="not-a-objectid")

        with self.assertRaisesMessage(ValidationError, "is not a valid Object Id."):
            ObjectIdModel.objects.create(field="not-a-objectid")


class SerializationTests(TestCase):
    test_data = (
        '[{"fields": {"field": "6754ed8e584bc9ceaae3c072"}, "model": '
        '"model_fields_.objectidmodel", "pk": null}]'
    )

    def test_dumping(self):
        instance = ObjectIdModel(field=ObjectId("6754ed8e584bc9ceaae3c072"))
        data = serializers.serialize("json", [instance])
        self.assertEqual(json.loads(data), json.loads(self.test_data))

    def test_loading(self):
        instance = next(serializers.deserialize("json", self.test_data)).object
        self.assertEqual(instance.field, ObjectId("6754ed8e584bc9ceaae3c072"))


class ValidationTests(TestCase):
    def test_invalid_objectid(self):
        field = ObjectIdField()
        msg = "“550e8400” is not a valid Object Id."
        with self.assertRaisesMessage(ValidationError, msg) as cm:
            field.clean("550e8400", None)
        self.assertEqual(cm.exception.code, "invalid")

    def test_objectid_instance_ok(self):
        value = ObjectId()
        field = ObjectIdField()
        self.assertEqual(field.clean(value, None), value)



"""
These tests are adapted from Django's tests/raw_query/tests.py.
"""

from datetime import date, datetime

from django.core.exceptions import FieldDoesNotExist
from django.db import connection
from django.test import TestCase

from django_mongodb_backend.queryset import RawQuerySet

from .models import (
    Author,
    Book,
    BookFkAsPk,
    Coffee,
    FriendlyAuthor,
    MixedCaseIDColumn,
    Reviewer,
)


class RawAggregateTests(TestCase):
    @classmethod
    def setUpTestData(cls):
        cls.a1 = Author.objects.create(first_name="Joe", last_name="Smith", dob=date(1950, 9, 20))
        cls.a2 = Author.objects.create(first_name="Jill", last_name="Doe", dob=date(1920, 4, 2))
        cls.a3 = Author.objects.create(first_name="Bob", last_name="Smith", dob=date(1986, 1, 25))
        cls.a4 = Author.objects.create(first_name="Bill", last_name="Jones", dob=date(1932, 5, 10))
        cls.b1 = Book.objects.create(
            title="The awesome book",
            author=cls.a1,
            paperback=False,
            opening_line=(
                "It was a bright cold day in April and the clocks were striking " "thirteen."
            ),
        )
        cls.b2 = Book.objects.create(
            title="The horrible book",
            author=cls.a1,
            paperback=True,
            opening_line=(
                "On an evening in the latter part of May a middle-aged man "
                "was walking homeward from Shaston to the village of Marlott, "
                "in the adjoining Vale of Blakemore, or Blackmoor."
            ),
        )
        cls.b3 = Book.objects.create(
            title="Another awesome book",
            author=cls.a1,
            paperback=False,
            opening_line="A squat gray building of only thirty-four stories.",
        )
        cls.b4 = Book.objects.create(
            title="Some other book",
            author=cls.a3,
            paperback=True,
            opening_line="It was the day my grandmother exploded.",
        )
        cls.c1 = Coffee.objects.create(brand="dunkin doughnuts")
        cls.c2 = Coffee.objects.create(brand="starbucks")
        cls.r1 = Reviewer.objects.create()
        cls.r2 = Reviewer.objects.create()
        cls.r1.reviewed.add(cls.b2, cls.b3, cls.b4)

    def assertSuccessfulRawQuery(
        self,
        model,
        query,
        expected_results,
        expected_annotations=(),
    ):
        """
        Execute the passed query against the passed model and check the output.
        """
        results = list(model.objects.raw_aggregate(query))
        expected_results = list(expected_results)
        with self.assertNumQueries(0):
            self.assertProcessed(model, results, expected_results, expected_annotations)
            self.assertAnnotations(results, expected_annotations)

    def assertProcessed(self, model, results, orig, expected_annotations=()):
        """Compare the results of a raw query against expected results."""
        self.assertEqual(len(results), len(orig))
        for index, item in enumerate(results):
            orig_item = orig[index]
            for annotation in expected_annotations:
                setattr(orig_item, *annotation)

            for field in model._meta.fields:
                # All values on the model are equal.
                self.assertEqual(getattr(item, field.attname), getattr(orig_item, field.attname))
                # This includes checking that they are the same type.
                self.assertEqual(
                    type(getattr(item, field.attname)),
                    type(getattr(orig_item, field.attname)),
                )

    def assertNoAnnotations(self, results):
        """The results of a raw query contain no annotations."""
        self.assertAnnotations(results, ())

    def assertAnnotations(self, results, expected_annotations):
        """The passed raw query results contain the expected annotations."""
        if expected_annotations:
            for index, result in enumerate(results):
                annotation, value = expected_annotations[index]
                self.assertTrue(hasattr(result, annotation))
                self.assertEqual(getattr(result, annotation), value)

    def test_rawqueryset_repr(self):
        queryset = RawQuerySet(pipeline=[])
        self.assertEqual(repr(queryset), "<RawQuerySet: []>")
        self.assertEqual(repr(queryset.query), "<RawQuery: []>")

    def test_simple_raw_query(self):
        """Basic test of raw query with a simple database query."""
        query = []
        authors = Author.objects.all()
        self.assertSuccessfulRawQuery(Author, query, authors)

    def test_raw_query_lazy(self):
        """
        Raw queries are lazy: they aren't actually executed until they're
        iterated over.
        """
        q = Author.objects.raw_aggregate([])
        self.assertIsNone(q.query.cursor)
        list(q)
        self.assertIsNotNone(q.query.cursor)

    def test_FK_raw_query(self):
        """
        Test of a simple raw query against a model containing a foreign key.
        """
        query = []
        books = Book.objects.all()
        self.assertSuccessfulRawQuery(Book, query, books)

    def test_db_column_handler(self):
        """
        Test of a simple raw query against a model containing a field with
        db_column defined.
        """
        query = []
        coffees = Coffee.objects.all()
        self.assertSuccessfulRawQuery(Coffee, query, coffees)

    def test_pk_with_mixed_case_db_column(self):
        """
        A raw query with a model that has a pk db_column with mixed case.
        """
        query = []
        queryset = MixedCaseIDColumn.objects.all()
        self.assertSuccessfulRawQuery(MixedCaseIDColumn, query, queryset)

    def test_order_handler(self):
        """
        Test of raw raw query's tolerance for columns being returned in any
        order.
        """
        selects = (
            ("dob, last_name, first_name, id"),
            ("last_name, dob, first_name, id"),
            ("first_name, last_name, dob, id"),
        )
        for select in selects:
            select = {col: 1 for col in select.split(", ")}
            query = [{"$project": select}]
            authors = Author.objects.all()
            self.assertSuccessfulRawQuery(Author, query, authors)

    def test_different_db_key_order(self):
        """
        A raw query correctly associates document keys to model fields when the
        document key order is different than the order of model fields.
        """
        author = Author(first_name="Out of", last_name="Order", dob=datetime(1950, 9, 20))
        # Insert a document with the keys reversed.
        connection.database.get_collection(Author._meta.db_table).insert_one(
            {
                field.name: getattr(author, field.name)
                for field in reversed(Author._meta.concrete_fields)
            }
        )
        query = []
        authors = Author.objects.all()
        self.assertSuccessfulRawQuery(Author, query, authors)

    def test_query_representation(self):
        """Test representation of raw query."""
        query = [{"$match": {"last_name": "foo"}}]
        qset = Author.objects.raw_aggregate(query)
        self.assertEqual(
            repr(qset),
            "<RawQuerySet: [{'$match': {'last_name': 'foo'}}]>",
        )
        self.assertEqual(
            repr(qset.query),
            "<RawQuery: [{'$match': {'last_name': 'foo'}}]>",
        )

    def test_many_to_many(self):
        """
        Test of a simple raw query against a model containing a m2m field.
        """
        query = []
        reviewers = Reviewer.objects.all()
        self.assertSuccessfulRawQuery(Reviewer, query, reviewers)

    def test_missing_fields(self):
        query = [{"$project": {"id": 1, "first_name": 1, "dob": 1}}]
        for author in Author.objects.raw_aggregate(query):
            self.assertIsNotNone(author.first_name)
            # last_name isn't given, but it will be retrieved on demand.
            self.assertIsNotNone(author.last_name)

    def test_missing_fields_without_PK(self):
        query = [{"$project": {"first_name": 1, "dob": 1, "_id": 0}}]
        msg = "Raw query must include the primary key"
        with self.assertRaisesMessage(FieldDoesNotExist, msg):
            list(Author.objects.raw_aggregate(query))

    def test_annotations(self):
        query = [
            {
                "$project": {
                    "first_name": 1,
                    "last_name": 1,
                    "dob": 1,
                    "birth_year": {"$year": "$dob"},
                },
            },
            {"$sort": {"_id": 1}},
        ]
        expected_annotations = (
            ("birth_year", 1950),
            ("birth_year", 1920),
            ("birth_year", 1986),
            ("birth_year", 1932),
        )
        authors = Author.objects.order_by("pk")
        self.assertSuccessfulRawQuery(Author, query, authors, expected_annotations)

    def test_multiple_iterations(self):
        query = []
        normal_authors = Author.objects.all()
        raw_authors = Author.objects.raw_aggregate(query)

        # First Iteration
        first_iterations = 0
        for index, raw_author in enumerate(raw_authors):
            self.assertEqual(normal_authors[index], raw_author)
            first_iterations += 1

        # Second Iteration
        second_iterations = 0
        for index, raw_author in enumerate(raw_authors):
            self.assertEqual(normal_authors[index], raw_author)
            second_iterations += 1

        self.assertEqual(first_iterations, second_iterations)

    def test_get_item(self):
        # Indexing on RawQuerySets
        query = [{"$sort": {"id": 1}}]
        third_author = Author.objects.raw_aggregate(query)[2]
        self.assertEqual(third_author.first_name, "Bob")

        first_two = Author.objects.raw_aggregate(query)[0:2]
        self.assertEqual(len(first_two), 2)

        with self.assertRaises(TypeError):
            Author.objects.raw_aggregate(query)["test"]

    def test_inheritance(self):
        f = FriendlyAuthor.objects.create(
            first_name="Wesley", last_name="Chun", dob=date(1962, 10, 28)
        )
        query = []
        self.assertEqual([o.pk for o in FriendlyAuthor.objects.raw_aggregate(query)], [f.pk])

    def test_query_count(self):
        self.assertNumQueries(1, list, Author.objects.raw_aggregate([]))

    def test_subquery_in_raw_sql(self):
        list(
            Book.objects.raw_aggregate(
                [{"$match": {"paperback": {"$ne": None}}}, {"$project": {"id": 1}}]
            )
        )

    def test_db_column_name_is_used_in_raw_query(self):
        """
        Regression test that ensures the `column` attribute on the field is
        used to generate the list of fields included in the query, as opposed
        to the `attname`. This is important when the primary key is a
        ForeignKey field because `attname` and `column` are not necessarily the
        same.
        """
        b = BookFkAsPk.objects.create(book=self.b1)
        self.assertEqual(
            list(
                BookFkAsPk.objects.raw_aggregate([{"$project": {"not_the_default": 1, "_id": 0}}])
            ),
            [b],
        )

    def test_result_caching(self):
        with self.assertNumQueries(1):
            books = Book.objects.raw_aggregate([])
            list(books)
            list(books)

    def test_iterator(self):
        with self.assertNumQueries(2):
            books = Book.objects.raw_aggregate([])
            list(books.iterator())
            list(books.iterator())

    def test_bool(self):
        self.assertIs(bool(Book.objects.raw_aggregate([])), True)
        self.assertIs(bool(Book.objects.raw_aggregate([{"$match": {"id": 0}}])), False)

    def test_len(self):
        self.assertEqual(len(Book.objects.raw_aggregate([])), 4)
        self.assertEqual(len(Book.objects.raw_aggregate([{"$match": {"id": 0}}])), 0)



import contextlib
import os

from django.core.exceptions import ImproperlyConfigured
from django.db.backends.base.base import BaseDatabaseWrapper
from django.utils.asyncio import async_unsafe
from django.utils.functional import cached_property
from pymongo.collection import Collection
from pymongo.driver_info import DriverInfo
from pymongo.mongo_client import MongoClient

from . import __version__ as django_mongodb_backend_version
from . import dbapi as Database
from .client import DatabaseClient
from .creation import DatabaseCreation
from .features import DatabaseFeatures
from .introspection import DatabaseIntrospection
from .operations import DatabaseOperations
from .query_utils import regex_match
from .schema import DatabaseSchemaEditor
from .utils import OperationDebugWrapper
from .validation import DatabaseValidation


class Cursor:
    """A "nodb" cursor that does nothing except work on a context manager."""

    def __enter__(self):
        pass

    def __exit__(self, exception_type, exception_value, exception_traceback):
        pass


class DatabaseWrapper(BaseDatabaseWrapper):
    data_types = {
        "AutoField": "int",
        "BigAutoField": "long",
        "BinaryField": "binData",
        "BooleanField": "bool",
        "CharField": "string",
        "DateField": "date",
        "DateTimeField": "date",
        "DecimalField": "decimal",
        "DurationField": "long",
        "FileField": "string",
        "FilePathField": "string",
        "FloatField": "double",
        "IntegerField": "int",
        "BigIntegerField": "long",
        "GenericIPAddressField": "string",
        "JSONField": "object",
        "OneToOneField": "int",
        "PositiveBigIntegerField": "int",
        "PositiveIntegerField": "long",
        "PositiveSmallIntegerField": "int",
        "SlugField": "string",
        "SmallAutoField": "int",
        "SmallIntegerField": "int",
        "TextField": "string",
        "TimeField": "date",
        "UUIDField": "string",
    }
    # Django uses these operators to generate SQL queries before it generates
    # MQL queries.
    operators = {
        "exact": "= %s",
        "iexact": "= UPPER(%s)",
        "contains": "LIKE %s",
        "icontains": "LIKE UPPER(%s)",
        "regex": "~ %s",
        "iregex": "~* %s",
        "gt": "> %s",
        "gte": ">= %s",
        "lt": "< %s",
        "lte": "<= %s",
        "startswith": "LIKE %s",
        "endswith": "LIKE %s",
        "istartswith": "LIKE UPPER(%s)",
        "iendswith": "LIKE UPPER(%s)",
    }
    # As with `operators`, these patterns are used to generate SQL before MQL.
    pattern_esc = "%%"
    pattern_ops = {
        "contains": "LIKE '%%' || {} || '%%'",
        "icontains": "LIKE '%%' || UPPER({}) || '%%'",
        "startswith": "LIKE {} || '%%'",
        "istartswith": "LIKE UPPER({}) || '%%'",
        "endswith": "LIKE '%%' || {}",
        "iendswith": "LIKE '%%' || UPPER({})",
    }

    def _isnull_operator(a, b):
        is_null = {
            "$or": [
                # The path does not exist (i.e. is "missing")
                {"$eq": [{"$type": a}, "missing"]},
                # or the value is None.
                {"$eq": [a, None]},
            ]
        }
        return is_null if b else {"$not": is_null}

    mongo_operators = {
        "exact": lambda a, b: {"$eq": [a, b]},
        "gt": lambda a, b: {"$gt": [a, b]},
        "gte": lambda a, b: {"$gte": [a, b]},
        # MongoDB considers null less than zero. Exclude null values to match
        # SQL behavior.
        "lt": lambda a, b: {"$and": [{"$lt": [a, b]}, {"$ne": [a, None]}]},
        "lte": lambda a, b: {"$and": [{"$lte": [a, b]}, {"$ne": [a, None]}]},
        "in": lambda a, b: {"$in": [a, b]},
        "isnull": _isnull_operator,
        "range": lambda a, b: {
            "$and": [
                {"$or": [DatabaseWrapper._isnull_operator(b[0], True), {"$gte": [a, b[0]]}]},
                {"$or": [DatabaseWrapper._isnull_operator(b[1], True), {"$lte": [a, b[1]]}]},
            ]
        },
        "iexact": lambda a, b: regex_match(a, ("^", b, {"$literal": "$"}), insensitive=True),
        "startswith": lambda a, b: regex_match(a, ("^", b)),
        "istartswith": lambda a, b: regex_match(a, ("^", b), insensitive=True),
        "endswith": lambda a, b: regex_match(a, (b, {"$literal": "$"})),
        "iendswith": lambda a, b: regex_match(a, (b, {"$literal": "$"}), insensitive=True),
        "contains": lambda a, b: regex_match(a, b),
        "icontains": lambda a, b: regex_match(a, b, insensitive=True),
        "regex": lambda a, b: regex_match(a, b),
        "iregex": lambda a, b: regex_match(a, b, insensitive=True),
    }

    display_name = "MongoDB"
    vendor = "mongodb"
    Database = Database
    SchemaEditorClass = DatabaseSchemaEditor
    client_class = DatabaseClient
    creation_class = DatabaseCreation
    features_class = DatabaseFeatures
    introspection_class = DatabaseIntrospection
    ops_class = DatabaseOperations
    validation_class = DatabaseValidation

    def get_collection(self, name, **kwargs):
        collection = Collection(self.database, name, **kwargs)
        if self.queries_logged:
            collection = OperationDebugWrapper(self, collection)
        return collection

    def get_database(self):
        if self.queries_logged:
            return OperationDebugWrapper(self)
        return self.database

    @cached_property
    def database(self):
        """Connect to the database the first time it's accessed."""
        if self.connection is None:
            self.connect()
        # Cache the database attribute set by init_connection_state()
        return self.database

    def init_connection_state(self):
        self.database = self.connection[self.settings_dict["NAME"]]
        super().init_connection_state()

    def get_connection_params(self):
        settings_dict = self.settings_dict
        if not settings_dict["NAME"]:
            raise ImproperlyConfigured('settings.DATABASES is missing the "NAME" value.')
        return {
            "host": settings_dict["HOST"] or None,
            "port": int(settings_dict["PORT"] or 27017),
            "username": settings_dict.get("USER"),
            "password": settings_dict.get("PASSWORD"),
            **settings_dict["OPTIONS"],
        }

    @async_unsafe
    def get_new_connection(self, conn_params):
        return MongoClient(**conn_params, driver=self._driver_info())

    def _driver_info(self):
        if not os.environ.get("RUNNING_DJANGOS_TEST_SUITE"):
            return DriverInfo("django-mongodb-backend", django_mongodb_backend_version)
        return None

    def _commit(self):
        pass

    def _rollback(self):
        pass

    def set_autocommit(self, autocommit, force_begin_transaction_with_broken_autocommit=False):
        self.autocommit = autocommit

    @async_unsafe
    def close(self):
        super().close()
        with contextlib.suppress(AttributeError):
            del self.database

    @async_unsafe
    def cursor(self):
        return Cursor()

    def get_database_version(self):
        """Return a tuple of the database's version."""
        return tuple(self.connection.server_info()["versionArray"])



from django.db import NotSupportedError
from django.db.models.fields.related_lookups import In, MultiColSource, RelatedIn
from django.db.models.lookups import (
    BuiltinLookup,
    FieldGetDbPrepValueIterableMixin,
    IsNull,
    PatternLookup,
    UUIDTextMixin,
)

from .query_utils import process_lhs, process_rhs


def builtin_lookup(self, compiler, connection):
    lhs_mql = process_lhs(self, compiler, connection)
    value = process_rhs(self, compiler, connection)
    return connection.mongo_operators[self.lookup_name](lhs_mql, value)


_field_resolve_expression_parameter = FieldGetDbPrepValueIterableMixin.resolve_expression_parameter


def field_resolve_expression_parameter(self, compiler, connection, sql, param):
    """For MongoDB, this method must call as_mql() instead of as_sql()."""
    sql, sql_params = _field_resolve_expression_parameter(self, compiler, connection, sql, param)
    if connection.vendor == "mongodb":
        params = [param]
        if hasattr(param, "resolve_expression"):
            param = param.resolve_expression(compiler.query)
        if hasattr(param, "as_mql"):
            params = [param.as_mql(compiler, connection)]
        return sql, params
    return sql, sql_params


def in_(self, compiler, connection):
    if isinstance(self.lhs, MultiColSource):
        raise NotImplementedError("MultiColSource is not supported.")
    db_rhs = getattr(self.rhs, "_db", None)
    if db_rhs is not None and db_rhs != connection.alias:
        raise ValueError(
            "Subqueries aren't allowed across different databases. Force "
            "the inner query to be evaluated using `list(inner_query)`."
        )
    return builtin_lookup(self, compiler, connection)


def get_subquery_wrapping_pipeline(self, compiler, connection, field_name, expr):  # noqa: ARG001
    return [
        {
            "$facet": {
                "group": [
                    {
                        "$group": {
                            "_id": None,
                            "tmp_name": {"$addToSet": expr.as_mql(compiler, connection)},
                        }
                    }
                ]
            }
        },
        {
            "$project": {
                field_name: {
                    "$ifNull": [
                        {
                            "$getField": {
                                "input": {"$arrayElemAt": ["$group", 0]},
                                "field": "tmp_name",
                            }
                        },
                        [],
                    ]
                }
            }
        },
    ]


def is_null(self, compiler, connection):
    if not isinstance(self.rhs, bool):
        raise ValueError("The QuerySet value for an isnull lookup must be True or False.")
    lhs_mql = process_lhs(self, compiler, connection)
    return connection.mongo_operators["isnull"](lhs_mql, self.rhs)


# from https://www.pcre.org/current/doc/html/pcre2pattern.html#SEC4
REGEX_MATCH_ESCAPE_CHARS = (
    ("\\", r"\\"),  # general escape character
    ("^", r"\^"),  # start of string
    ({"$literal": "$"}, r"\$"),  # end of string
    (".", r"\."),  # match any character
    ("[", r"\["),  # start character class definition
    ("|", r"\|"),  # start of alternative branch
    ("(", r"\("),  # start group or control verb
    (")", r"\)"),  # end group or control verb
    ("*", r"\*"),  #  0 or more quantifier
    ("+", r"\+"),  #  1 or more quantifier
    ("?", r"\?"),  # 0 or 1 quantifier
    ("{", r"\}"),  # start min/max quantifier
)


def pattern_lookup_prep_lookup_value(self, value):
    if hasattr(self.rhs, "as_mql"):
        # If value is a column reference, escape $regexMatch special chars.
        # Analogous to PatternLookup.get_rhs_op() / pattern_esc.
        for find, replacement in REGEX_MATCH_ESCAPE_CHARS:
            value = {"$replaceAll": {"input": value, "find": find, "replacement": replacement}}
    else:
        # If value is a literal, remove percent signs added by
        # PatternLookup.process_rhs() for LIKE queries.
        if self.lookup_name in ("startswith", "istartswith"):
            value = value[:-1]
        elif self.lookup_name in ("endswith", "iendswith"):
            value = value[1:]
        elif self.lookup_name in ("contains", "icontains"):
            value = value[1:-1]
    return value


def uuid_text_mixin(self, compiler, connection):  # noqa: ARG001
    raise NotSupportedError("Pattern lookups on UUIDField are not supported.")


def register_lookups():
    BuiltinLookup.as_mql = builtin_lookup
    FieldGetDbPrepValueIterableMixin.resolve_expression_parameter = (
        field_resolve_expression_parameter
    )
    In.as_mql = RelatedIn.as_mql = in_
    In.get_subquery_wrapping_pipeline = get_subquery_wrapping_pipeline
    IsNull.as_mql = is_null
    PatternLookup.prep_lookup_value_mongo = pattern_lookup_prep_lookup_value
    UUIDTextMixin.as_mql = uuid_text_mixin



from django.db.models.fields import DurationField

_get_db_prep_value = DurationField.get_db_prep_value


def get_db_prep_value(self, value, connection, prepared=False):
    """DurationField stores milliseconds rather than microseconds."""
    value = _get_db_prep_value(self, value, connection, prepared)
    if connection.vendor == "mongodb" and value is not None:
        value //= 1000
    return value


def register_duration_field():
    DurationField.get_db_prep_value = get_db_prep_value
